/*
    AutoX11, a tool to automate tasks in an X11 environment
    Copyright (C) 2014  Alicia <alicia@ion.nu>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU Affero General Public License as published by
    the Free Software Foundation, version 3 of the License.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Affero General Public License for more details.

    You should have received a copy of the GNU Affero General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <stdio.h>
#include <unistd.h>
#include <fcntl.h>
#include <stdlib.h>
#include <string.h>
#include <wchar.h>
#include <locale.h>
#include <X11/Xlib.h>
#include <X11/extensions/XTest.h>
#include <X11/Xutil.h>
#include <gdk-pixbuf/gdk-pixbuf.h>
#include "autox11.h"

const char* helptext="Usage: autox11 <action> [<options>]\n"
"\n"
"Valid actions and options:\n"
"clickon <sample image>\n"
"   -x, --xoffset = where on the matched area to click, default=middle\n"
"   -y, --yoffset = where on the matched area to click, default=middle\n"
"   -w, --wait    = don't return until the target has been found and clicked\n"
"   -b, --button  = set button to click with, default=1 (left)\n"
"waitfor <sample image>\n"
"   (just wait for the sample image to appear somewhere on-screen)\n"
"type <text>\n"
"key <name of key>\n"
"   -p, --press   = press and hold the key\n"
"   -r, --release = release the key\n";

int main(int argc, char** argv)
{
  if(argc<2)
  {
    printf("%s", helptext);
    return 1;
  }
  int i;
  for(i=1; i<argc; ++i)
  {
    if(!strcmp(argv[i], "-h") || !strcmp(argv[i], "--help")){printf("%s", helptext); return 0;}
  }
  Display* display=XOpenDisplay(getenv("DISPLAY"));
  if(!display)
  {
    printf("Failed to open X11 display\n");
    return 1;
  }
  int ret=0;
  int screen=XDefaultScreen(display);
  if(!strcmp(argv[1], "clickon"))
  {
    ret=2; // Return non-zero if we didn't find the target
    int xo=-1;
    int yo=-1;
    const char* target=0;
    char cwait=0;
    int button=1;
    for(i=2; i<argc; ++i)
    {
      if(!strcmp(argv[i], "-x")||!strcmp(argv[i], "--xoffset")){++i; xo=atoi(argv[i]);}
      else if(!strcmp(argv[i], "-y")||!strcmp(argv[i], "--yoffset")){++i; yo=atoi(argv[i]);}
      else if(!strcmp(argv[i], "-w")||!strcmp(argv[i], "--wait")){cwait=1;}
      else if(!strcmp(argv[i], "-b")||!strcmp(argv[i], "--button")){++i; button=atoi(argv[i]);}
      else if(argv[i][0]!='-'){target=argv[i];}
      else{printf("Unknown option '%s'\n", argv[i]); XCloseDisplay(display); return 1;}
    }
    if(!target){printf("Pick a target\n"); XCloseDisplay(display); return 1;}
    GdkPixbuf* pic=gdk_pixbuf_new_from_file(target, 0);
    if(!pic){printf("Failed to load '%s'\n", target); XCloseDisplay(display); return 1;}
    while(1)
    {
      XImage* img=getscreenshot(display, screen);
      void pointat(int x, int y)
      {
        cwait=0; // Found it, stop waiting
        ret=0;
        if(xo==-1){xo=gdk_pixbuf_get_width(pic)/2;}
        if(yo==-1){yo=gdk_pixbuf_get_height(pic)/2;}
        XTestFakeMotionEvent(display, screen, x+xo, y+yo, 0);
        XTestFakeButtonEvent(display, button, 1, 0);
        XTestFakeButtonEvent(display, button, 0, 0);
      }
      findneedle(img, pic, pointat);
      XDestroyImage(img);
      if(!cwait){break;}
      usleep(100000); // 0.1 seconds seems like a reasonable compromise between "check as often as possible" and "don't use too much CPU time"
    }
    g_object_unref(pic);
  }
  else if(!strcmp(argv[1], "waitfor"))
  {
    char cwait=1;
    const char* target=0;
    for(i=2; i<argc; ++i)
    {
      if(argv[i][0]!='-'){target=argv[i];}
      else{printf("Unknown option '%s'\n", argv[i]); XCloseDisplay(display); return 1;}
    }
    if(!target){printf("Pick a target\n"); XCloseDisplay(display); return 1;}
    GdkPixbuf* pic=gdk_pixbuf_new_from_file(target, 0);
    if(!pic){printf("Failed to load '%s'\n", target); XCloseDisplay(display); return 1;}
    while(1)
    {
      XImage* img=getscreenshot(display, screen);
      void found(int x, int y)
      {
        cwait=0; // Found it, stop waiting
      }
      findneedle(img, pic, found);
      XDestroyImage(img);
      if(!cwait){break;}
      usleep(100000);
    }
    g_object_unref(pic);
  }
  else if(!strcmp(argv[1], "type"))
  {
    setlocale(LC_ALL, "");
    int j;
    for(i=2; i<argc; ++i)
    {
      if(i>2){typechar(display, L' ');}
      int len=mbstowcs(0, argv[i], 0);
      wchar_t str[len+1];
      mbstowcs(str, argv[i], len+1);
      for(j=0; str[j]; ++j)
      {
        typechar(display, str[j]);
      }
    }
  }
  else if(!strcmp(argv[1], "key"))
  {
    const char* key;
    char press=1;
    char release=1;
    for(i=2; i<argc; ++i)
    {
      if(!strcmp(argv[i], "-p")||!strcmp(argv[i], "--press")){press=1; release=0;}
      else if(!strcmp(argv[i], "-r")||!strcmp(argv[i], "--release")){release=1; press=0;}
      else if(argv[i][0]!='-'){key=argv[i];}
      else{printf("Unknown option '%s'\n", argv[i]); XCloseDisplay(display); return 1;}
    }
    if(!key){printf("Pick a key\n"); XCloseDisplay(display); return 1;}
    KeySym sym=XStringToKeysym(key);
    if(!sym){printf("Key '%s' not found\n", key); guesskey(key); XCloseDisplay(display); return 1;}
    KeyCode keycode=XKeysymToKeycode(display, sym);
    if(press){XTestFakeKeyEvent(display, keycode, 1, 0);}
    if(release){XTestFakeKeyEvent(display, keycode, 0, 0);}
  }
  else{printf("Unknown action '%s'\n", argv[1]); ret=1;}
  XCloseDisplay(display);
  return ret;
}
